@@ -10,6 +10,18 @@
 
 
 #include <cuda.h>
+
+#ifdef CUDAPP_PRETEND_CUDA_VERSION
+#define CUDAPP_CUDA_VERSION CUDAPP_PRETEND_CUDA_VERSION
+#else
+#define CUDAPP_CUDA_VERSION CUDA_VERSION
+#endif
+
+#if CUDAPP_CUDA_VERSION >= 4000
+#include <cudaProfiler.h>
+#endif
+
+#include <stdint.h>
 #include <stdexcept>
 #include <boost/shared_ptr.hpp>
 #include <boost/foreach.hpp>
@@ -28,12 +40,42 @@
 #warning *****************************************************************
 #endif
 
+// MAYBE? cuMemcpy, cuPointerGetAttribute
+// TODO: cuCtxSetCurrent, cuCtxGetCurrent 
+// (use once the old, deprecated functions have been removed from CUDA)
+
 
 
 
 // #define CUDAPP_TRACE_CUDA
 #define CUDAPP_POST_30_BETA
 
+#ifdef CUDAPP_PRETEND_CUDA_VERSION
+#define CUDAPP_CUDA_VERSION CUDAPP_PRETEND_CUDA_VERSION
+#else
+#define CUDAPP_CUDA_VERSION CUDA_VERSION
+#endif
+
+
+
+
+#if PY_VERSION_HEX >= 0x02050000
+  typedef Py_ssize_t PYCUDA_BUFFER_SIZE_T;
+#else
+  typedef int PYCUDA_BUFFER_SIZE_T;
+#endif
+
+
+
+#define PYCUDA_PARSE_STREAM_PY \
+    CUstream s_handle; \
+    if (stream_py.ptr() != Py_None) \
+    { \
+      const stream &s = py::extract<const stream &>(stream_py); \
+      s_handle = s.handle(); \
+    } \
+    else \
+      s_handle = 0;
 
 
 
@@ -61,7 +103,7 @@
       cu_status_code = NAME ARGLIST; \
     Py_END_ALLOW_THREADS \
     if (cu_status_code != CUDA_SUCCESS) \
-      throw cuda::error(#NAME, cu_status_code);\
+      throw pycuda::error(#NAME, cu_status_code);\
   }
 
 #define CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(NAME, ARGLIST, TRACE_INFO) \
@@ -71,7 +113,7 @@
     cu_status_code = NAME ARGLIST; \
     CUDAPP_PRINT_ERROR_TRACE(#NAME, cu_status_code); \
     if (cu_status_code != CUDA_SUCCESS) \
-      throw cuda::error(#NAME, cu_status_code);\
+      throw pycuda::error(#NAME, cu_status_code);\
   }
 
 #define CUDAPP_CALL_GUARDED_THREADED(NAME, ARGLIST) \
@@ -83,7 +125,7 @@
     Py_END_ALLOW_THREADS \
     CUDAPP_PRINT_ERROR_TRACE(#NAME, cu_status_code); \
     if (cu_status_code != CUDA_SUCCESS) \
-      throw cuda::error(#NAME, cu_status_code);\
+      throw pycuda::error(#NAME, cu_status_code);\
   }
 
 #define CUDAPP_CALL_GUARDED(NAME, ARGLIST) \
@@ -93,7 +135,7 @@
     cu_status_code = NAME ARGLIST; \
     CUDAPP_PRINT_ERROR_TRACE(#NAME, cu_status_code); \
     if (cu_status_code != CUDA_SUCCESS) \
-      throw cuda::error(#NAME, cu_status_code);\
+      throw pycuda::error(#NAME, cu_status_code);\
   }
 #define CUDAPP_CALL_GUARDED_CLEANUP(NAME, ARGLIST) \
   { \
@@ -105,13 +147,13 @@
       std::cerr \
         << "PyCUDA WARNING: a clean-up operation failed (dead context maybe?)" \
         << std::endl \
-        << cuda::error::make_message(#NAME, cu_status_code) \
+        << pycuda::error::make_message(#NAME, cu_status_code) \
         << std::endl; \
   }
 #define CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(TYPE) \
-  catch (cuda::cannot_activate_out_of_thread_context) \
+  catch (pycuda::cannot_activate_out_of_thread_context) \
   { } \
-  catch (cuda::cannot_activate_dead_context) \
+  catch (pycuda::cannot_activate_dead_context) \
   { \
     /* PyErr_Warn( \
         PyExc_UserWarning, #TYPE " in dead context was implicitly cleaned up");*/ \
@@ -125,18 +167,26 @@
 
 
 
-namespace cuda
+namespace pycuda
 {
   namespace py = boost::python;
 
   typedef
-#if CUDA_VERSION >= 3020
+#if CUDAPP_CUDA_VERSION >= 3020
         size_t
 #else
         unsigned int 
 #endif
         pycuda_size_t;
 
+  typedef 
+#if defined(_WIN32) && defined(_WIN64)
+    long long
+#else
+    long
+#endif
+    hash_type;
+
 
 
   // {{{ error reporting
@@ -175,6 +225,11 @@ namespace cuda
         return m_code;
       }
 
+      bool is_out_of_memory() const
+      {
+        return code() == CUDA_ERROR_OUT_OF_MEMORY;
+      }
+
       static const char *curesult_to_str(CUresult e)
       {
         switch (e)
@@ -183,9 +238,15 @@ namespace cuda
           case CUDA_ERROR_INVALID_VALUE: return "invalid value";
           case CUDA_ERROR_OUT_OF_MEMORY: return "out of memory";
           case CUDA_ERROR_NOT_INITIALIZED: return "not initialized";
-#if CUDA_VERSION >= 2000
+#if CUDAPP_CUDA_VERSION >= 2000
           case CUDA_ERROR_DEINITIALIZED: return "deinitialized";
 #endif
+#if CUDAPP_CUDA_VERSION >= 4000
+          case CUDA_ERROR_PROFILER_DISABLED: return "profiler disabled";
+          case CUDA_ERROR_PROFILER_NOT_INITIALIZED: return "profiler not initialized";
+          case CUDA_ERROR_PROFILER_ALREADY_STARTED: return "profiler already started";
+          case CUDA_ERROR_PROFILER_ALREADY_STOPPED: return "profiler already stopped";
+#endif
 
           case CUDA_ERROR_NO_DEVICE: return "no device";
           case CUDA_ERROR_INVALID_DEVICE: return "invalid device";
@@ -200,20 +261,23 @@ namespace cuda
           case CUDA_ERROR_NO_BINARY_FOR_GPU: return "no binary for gpu";
           case CUDA_ERROR_ALREADY_ACQUIRED: return "already acquired";
           case CUDA_ERROR_NOT_MAPPED: return "not mapped";
-#if CUDA_VERSION >= 3000
+#if CUDAPP_CUDA_VERSION >= 3000
           case CUDA_ERROR_NOT_MAPPED_AS_ARRAY: return "not mapped as array";
           case CUDA_ERROR_NOT_MAPPED_AS_POINTER: return "not mapped as pointer";
 #ifdef CUDAPP_POST_30_BETA
           case CUDA_ERROR_ECC_UNCORRECTABLE: return "ECC uncorrectable";
 #endif
 #endif
-#if CUDA_VERSION >= 3010
+#if CUDAPP_CUDA_VERSION >= 3010
           case CUDA_ERROR_UNSUPPORTED_LIMIT: return "unsupported limit";
 #endif
+#if CUDAPP_CUDA_VERSION >= 4000
+          case CUDA_ERROR_CONTEXT_ALREADY_IN_USE: return "context already in use";
+#endif
 
           case CUDA_ERROR_INVALID_SOURCE: return "invalid source";
           case CUDA_ERROR_FILE_NOT_FOUND: return "file not found";
-#if CUDA_VERSION >= 3010
+#if CUDAPP_CUDA_VERSION >= 3010
           case CUDA_ERROR_SHARED_OBJECT_SYMBOL_NOT_FOUND:
             return "shared object symbol not found";
           case CUDA_ERROR_SHARED_OBJECT_INIT_FAILED:
@@ -231,7 +295,14 @@ namespace cuda
           case CUDA_ERROR_LAUNCH_TIMEOUT: return "launch timeout";
           case CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING: return "launch incompatible texturing";
 
-#if (CUDA_VERSION >= 3000) && (CUDA_VERSION < 3020)
+#if CUDAPP_CUDA_VERSION >= 4000
+          case CUDA_ERROR_PEER_ACCESS_ALREADY_ENABLED: return "peer access already enabled";
+          case CUDA_ERROR_PEER_ACCESS_NOT_ENABLED: return "peer access not enabled";
+          case CUDA_ERROR_PRIMARY_CONTEXT_ACTIVE: return "primary context active";
+          case CUDA_ERROR_CONTEXT_IS_DESTROYED: return "context is destroyed";
+#endif
+
+#if (CUDAPP_CUDA_VERSION >= 3000) && (CUDAPP_CUDA_VERSION < 3020)
           case CUDA_ERROR_POINTER_IS_64BIT:
              return "attempted to retrieve 64-bit pointer via 32-bit api function";
           case CUDA_ERROR_SIZE_IS_64BIT:
@@ -262,7 +333,7 @@ namespace cuda
   // }}}
 
   // {{{ version query ------------------------------------------------------------
-#if CUDA_VERSION >= 2020
+#if CUDAPP_CUDA_VERSION >= 2020
   inline int get_driver_version()
   {
     int result;
@@ -331,7 +402,7 @@ namespace cuda
         return m_device != other.m_device;
       }
 
-      long hash() const
+      hash_type hash() const
       {
         return m_device;
       }
@@ -341,6 +412,15 @@ namespace cuda
       CUdevice handle() const
       { return m_device; }
 
+#if CUDAPP_CUDA_VERSION >= 4000
+      bool can_access_peer(device const &other)
+      {
+        int result;
+        CUDAPP_CALL_GUARDED(cuDeviceCanAccessPeer, (&result, handle(), other.handle()));
+        return result;
+      }
+#endif
+
   };
 
   inline
@@ -450,9 +530,9 @@ namespace cuda
         return m_context != other.m_context;
       }
 
-      long hash() const
+      hash_type hash() const
       {
-        return long(m_context) ^ long(this);
+        return hash_type(m_context) ^ hash_type(this);
       }
 
       boost::thread::id thread_id() const
@@ -513,7 +593,7 @@ namespace cuda
         return device(dev);
       }
 
-#if CUDA_VERSION >= 2000
+#if CUDAPP_CUDA_VERSION >= 2000
 
       static void prepare_context_switch()
       {
@@ -572,7 +652,7 @@ namespace cuda
         }
       }
 
-#if CUDA_VERSION >= 3010
+#if CUDAPP_CUDA_VERSION >= 3010
       static void set_limit(CUlimit limit, size_t value)
       {
         CUDAPP_CALL_GUARDED(cuCtxSetLimit, (limit, value));
@@ -586,7 +666,7 @@ namespace cuda
       }
 #endif
 
-#if CUDA_VERSION >= 3020
+#if CUDAPP_CUDA_VERSION >= 3020
       static CUfunc_cache get_cache_config()
       {
         CUfunc_cache value;
@@ -607,6 +687,18 @@ namespace cuda
       }
 #endif
 
+#if CUDAPP_CUDA_VERSION >= 4000
+      static void enable_peer_access(context const &peer, unsigned int flags)
+      {
+        CUDAPP_CALL_GUARDED(cuCtxEnablePeerAccess, (peer.handle(), flags));
+      }
+
+      static void disable_peer_access(context const &peer)
+      {
+        CUDAPP_CALL_GUARDED(cuCtxDisablePeerAccess, (peer.handle()));
+      }
+#endif
+
       friend class device;
       friend void context_push(boost::shared_ptr<context> ctx);
       friend boost::shared_ptr<context>
@@ -631,7 +723,7 @@ namespace cuda
 
 
 
-#if CUDA_VERSION >= 2000
+#if CUDAPP_CUDA_VERSION >= 2000
   inline
   void context_push(boost::shared_ptr<context> ctx)
   {
@@ -715,19 +807,19 @@ namespace cuda
         : m_context(ctx)
       {
         if (!m_context->is_valid())
-          throw cuda::cannot_activate_dead_context(
+          throw pycuda::cannot_activate_dead_context(
               "cannot activate dead context");
 
         m_did_switch = context::current_context() != m_context;
         if (m_did_switch)
         {
           if (boost::this_thread::get_id() != m_context->thread_id())
-            throw cuda::cannot_activate_out_of_thread_context(
+            throw pycuda::cannot_activate_out_of_thread_context(
                 "cannot activate out-of-thread context");
-#if CUDA_VERSION >= 2000
+#if CUDAPP_CUDA_VERSION >= 2000
           context_push(m_context);
 #else
-          throw cuda::error("scoped_context_activation", CUDA_ERROR_INVALID_CONTEXT,
+          throw pycuda::error("scoped_context_activation", CUDA_ERROR_INVALID_CONTEXT,
               "not available in CUDA < 2.0");
 #endif
         }
@@ -735,7 +827,7 @@ namespace cuda
 
       ~scoped_context_activation()
       {
-#if CUDA_VERSION >= 2000
+#if CUDAPP_CUDA_VERSION >= 2000
         if (m_did_switch)
           m_context->pop();
 #endif
@@ -746,6 +838,8 @@ namespace cuda
   // }}}
 
   // {{{ stream
+  class event;
+
   class stream : public boost::noncopyable, public context_dependent
   {
     private:
@@ -771,6 +865,13 @@ namespace cuda
       CUstream handle() const
       { return m_stream; }
 
+      intptr_t handle_int() const
+      { return (intptr_t) m_stream; }
+
+#if CUDAPP_CUDA_VERSION >= 3020
+      void wait_for_event(const event &evt);
+#endif
+
       bool is_done() const
       {
         CUDAPP_PRINT_CALL_TRACE("cuStreamQuery");
@@ -802,7 +903,7 @@ namespace cuda
         : m_managed(true)
       { CUDAPP_CALL_GUARDED(cuArrayCreate, (&m_array, &descr)); }
 
-#if CUDA_VERSION >= 2000
+#if CUDAPP_CUDA_VERSION >= 2000
       array(const CUDA_ARRAY3D_DESCRIPTOR &descr)
         : m_managed(true)
       { CUDAPP_CALL_GUARDED(cuArray3DCreate, (&m_array, &descr)); }
@@ -838,7 +939,7 @@ namespace cuda
         return result;
       }
 
-#if CUDA_VERSION >= 2000
+#if CUDAPP_CUDA_VERSION >= 2000
       CUDA_ARRAY3D_DESCRIPTOR get_descriptor_3d()
       {
         CUDA_ARRAY3D_DESCRIPTOR result;
@@ -903,14 +1004,14 @@ namespace cuda
               m_texref, dptr, bytes));
 
         if (!allow_offset && byte_offset != 0)
-          throw cuda::error("texture_reference::set_address", CUDA_ERROR_INVALID_VALUE,
+          throw pycuda::error("texture_reference::set_address", CUDA_ERROR_INVALID_VALUE,
               "texture binding resulted in offset, but allow_offset was false");
 
         m_array.reset();
         return byte_offset;
       }
 
-#if CUDA_VERSION >= 2020
+#if CUDAPP_CUDA_VERSION >= 2020
       void set_address_2d(CUdeviceptr dptr,
           const CUDA_ARRAY_DESCRIPTOR &descr, unsigned int pitch)
       {
@@ -954,7 +1055,7 @@ namespace cuda
         return result;
       }
 
-#if CUDA_VERSION >= 2000
+#if CUDAPP_CUDA_VERSION >= 2000
       py::tuple get_format()
       {
         CUarray_format fmt;
@@ -975,7 +1076,7 @@ namespace cuda
   // }}}
 
   // {{{ surface reference
-#if CUDA_VERSION >= 3010
+#if CUDAPP_CUDA_VERSION >= 3010
   class module;
 
   class surface_reference : public  boost::noncopyable
@@ -1071,7 +1172,7 @@ namespace cuda
     return result.release();
   }
 
-#if CUDA_VERSION >= 3010
+#if CUDAPP_CUDA_VERSION >= 3010
   inline
   surface_reference *module_get_surfref(
       boost::shared_ptr<module> mod, const char *name)
@@ -1125,10 +1226,11 @@ namespace cuda
         CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
           cuParamSetf, (m_function, offset, value), m_symbol);
       }
-      void param_setv(int offset, void *buf, unsigned long len)
+      void param_setv(int offset, void *buf, size_t len)
       {
+        // maybe the unsigned int will change, it does not seem right
         CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
-            cuParamSetv, (m_function, offset, buf, len), m_symbol);
+          cuParamSetv, (m_function, offset, buf, (unsigned int) len), m_symbol);
       }
       void param_set_texref(const texture_reference &tr)
       {
@@ -1153,7 +1255,7 @@ namespace cuda
             m_symbol);
       }
 
-#if CUDA_VERSION >= 2020
+#if CUDAPP_CUDA_VERSION >= 2020
       int get_attribute(CUfunction_attribute attr) const
       {
         int result;
@@ -1163,13 +1265,67 @@ namespace cuda
       }
 #endif
 
-#if CUDA_VERSION >= 3000 && defined(CUDAPP_POST_30_BETA)
+#if CUDAPP_CUDA_VERSION >= 3000 && defined(CUDAPP_POST_30_BETA)
       void set_cache_config(CUfunc_cache fc)
       {
         CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
             cuFuncSetCacheConfig, (m_function, fc), m_symbol);
       }
 #endif
+
+#if CUDAPP_CUDA_VERSION >= 4000
+      void launch_kernel(py::tuple grid_dim_py, py::tuple block_dim_py,
+          py::object parameter_buffer,
+          unsigned shared_mem_bytes, py::object stream_py)
+      {
+        const unsigned axis_count = 3;
+        unsigned grid_dim[axis_count];
+        unsigned block_dim[axis_count];
+
+        for (unsigned i = 0; i < axis_count; ++i)
+        {
+          grid_dim[i] = 1;
+          block_dim[i] = 1;
+        }
+
+        unsigned gd_length = py::len(grid_dim_py);
+        if (gd_length > axis_count)
+          throw pycuda::error("function::launch_kernel", CUDA_ERROR_INVALID_HANDLE,
+              "too many grid dimensions in kernel launch");
+
+        for (unsigned i = 0; i < gd_length; ++i)
+          grid_dim[i] = py::extract<unsigned>(grid_dim_py[i]);
+
+        unsigned bd_length = py::len(block_dim_py);
+        if (bd_length > axis_count)
+          throw pycuda::error("function::launch_kernel", CUDA_ERROR_INVALID_HANDLE,
+              "too many block dimensions in kernel launch");
+
+        for (unsigned i = 0; i < bd_length; ++i)
+          block_dim[i] = py::extract<unsigned>(block_dim_py[i]);
+
+        PYCUDA_PARSE_STREAM_PY;
+
+        const void *par_buf;
+        PYCUDA_BUFFER_SIZE_T py_par_len;
+        if (PyObject_AsReadBuffer(parameter_buffer.ptr(), &par_buf, &py_par_len))
+          throw py::error_already_set();
+        size_t par_len = py_par_len;
+
+        void *config[] = {
+          CU_LAUNCH_PARAM_BUFFER_POINTER, const_cast<void *>(par_buf),
+          CU_LAUNCH_PARAM_BUFFER_SIZE, &par_len,
+          CU_LAUNCH_PARAM_END
+        };
+
+        CUDAPP_CALL_GUARDED(
+            cuLaunchKernel, (m_function, 
+              grid_dim[0], grid_dim[1], grid_dim[2],
+              block_dim[0], block_dim[1], block_dim[2],
+              shared_mem_bytes, s_handle, 0, config
+              ));
+      }
+#endif
   };
 
   inline
@@ -1192,7 +1348,7 @@ namespace cuda
   }
 
   inline
-  CUdeviceptr mem_alloc(unsigned long bytes)
+  CUdeviceptr mem_alloc(size_t bytes)
   {
     CUdeviceptr devptr;
     CUDAPP_CALL_GUARDED(cuMemAlloc, (&devptr, bytes));
@@ -1245,7 +1401,7 @@ namespace cuda
           m_valid = false;
         }
         else
-          throw cuda::error("device_allocation::free", CUDA_ERROR_INVALID_HANDLE);
+          throw pycuda::error("device_allocation::free", CUDA_ERROR_INVALID_HANDLE);
       }
 
       ~device_allocation()
@@ -1258,7 +1414,7 @@ namespace cuda
       { return m_devptr; }
   };
 
-  inline unsigned int mem_alloc_pitch(
+  inline Py_ssize_t mem_alloc_pitch(
       std::auto_ptr<device_allocation> &da,
         unsigned int width, unsigned int height, unsigned int access_size)
   {
@@ -1278,8 +1434,6 @@ namespace cuda
     return py::make_tuple(base, size);
   }
 
-  // missing: htoa, atoh, dtoh, htod
-
   inline
   void memcpy_dtoa(array const &ary, unsigned int index, CUdeviceptr src, unsigned int len)
   { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyDtoA, (ary.handle(), index, src, len)); }
@@ -1298,11 +1452,6 @@ namespace cuda
   // }}}
 
   // {{{ structured memcpy
-#if PY_VERSION_HEX >= 0x02050000
-  typedef Py_ssize_t PYCUDA_BUFFER_SIZE_T;
-#else
-  typedef int PYCUDA_BUFFER_SIZE_T;
-#endif
 
 #define MEMCPY_SETTERS \
     void set_src_host(py::object buf_py) \
@@ -1345,6 +1494,27 @@ namespace cuda
       dstDevice = devptr; \
     }
 
+#if CUDAPP_CUDA_VERSION >= 4000
+#define MEMCPY_SETTERS_UNIFIED \
+    void set_src_unified(py::object buf_py) \
+    { \
+      srcMemoryType = CU_MEMORYTYPE_UNIFIED; \
+      PYCUDA_BUFFER_SIZE_T len; \
+      if (PyObject_AsReadBuffer(buf_py.ptr(), &srcHost, &len)) \
+        throw py::error_already_set(); \
+    } \
+    \
+    void set_dst_unified(py::object buf_py) \
+    { \
+      dstMemoryType = CU_MEMORYTYPE_UNIFIED; \
+      PYCUDA_BUFFER_SIZE_T len; \
+      if (PyObject_AsWriteBuffer(buf_py.ptr(), &dstHost, &len)) \
+        throw py::error_already_set(); \
+    }
+#else
+#define MEMCPY_SETTERS_UNIFIED /* empty */
+#endif
+
 
 
 
@@ -1361,8 +1531,9 @@ namespace cuda
     }
 
     MEMCPY_SETTERS;
+    MEMCPY_SETTERS_UNIFIED;
 
-    void execute(bool aligned) const
+    void execute(bool aligned=false) const
     {
       if (aligned)
       { CUDAPP_CALL_GUARDED_THREADED(cuMemcpy2D, (this)); }
@@ -1374,7 +1545,7 @@ namespace cuda
     { CUDAPP_CALL_GUARDED_THREADED(cuMemcpy2DAsync, (this, s.handle())); }
   };
 
-#if CUDA_VERSION >= 2000
+#if CUDAPP_CUDA_VERSION >= 2000
   struct memcpy_3d : public CUDA_MEMCPY3D
   {
     memcpy_3d()
@@ -1394,6 +1565,7 @@ namespace cuda
     }
 
     MEMCPY_SETTERS;
+    MEMCPY_SETTERS_UNIFIED;
 
     void execute() const
     {
@@ -1405,48 +1577,205 @@ namespace cuda
   };
 #endif
 
+#if CUDAPP_CUDA_VERSION >= 4000
+  struct memcpy_3d_peer : public CUDA_MEMCPY3D_PEER
+  {
+    memcpy_3d_peer()
+    {
+      srcXInBytes = 0;
+      srcY = 0;
+      srcZ = 0;
+      srcLOD = 0;
+
+      dstXInBytes = 0;
+      dstY = 0;
+      dstZ = 0;
+      dstLOD = 0;
+    }
+
+    MEMCPY_SETTERS;
+    MEMCPY_SETTERS_UNIFIED;
+
+    void set_src_context(context const &ctx)
+    {
+      srcContext = ctx.handle();
+    }
+
+    void set_dst_context(context const &ctx)
+    {
+      dstContext = ctx.handle();
+    }
+
+    void execute() const
+    {
+      CUDAPP_CALL_GUARDED_THREADED(cuMemcpy3DPeer, (this));
+    }
+
+    void execute_async(const stream &s) const
+    { CUDAPP_CALL_GUARDED_THREADED(cuMemcpy3DPeerAsync, (this, s.handle())); }
+  };
+#endif
+
   // }}}
 
   // {{{ host memory
-  inline void *mem_alloc_host(unsigned int size, unsigned flags=0)
+  inline void *mem_host_alloc(size_t size, unsigned flags=0)
   {
     void *m_data;
-#if CUDA_VERSION >= 2020
+#if CUDAPP_CUDA_VERSION >= 2020
     CUDAPP_CALL_GUARDED(cuMemHostAlloc, (&m_data, size, flags));
 #else
     if (flags != 0)
-      throw cuda::error("mem_alloc_host", CUDA_ERROR_INVALID_VALUE,
-          "nonzero flags in mem_alloc_host not allowed in CUDA 2.1 and older");
+      throw pycuda::error("mem_host_alloc", CUDA_ERROR_INVALID_VALUE,
+          "nonzero flags in mem_host_alloc not allowed in CUDA 2.1 and older");
     CUDAPP_CALL_GUARDED(cuMemAllocHost, (&m_data, size));
 #endif
     return m_data;
   }
 
-  inline void mem_free_host(void *ptr)
+  inline void mem_host_free(void *ptr)
   {
     CUDAPP_CALL_GUARDED_CLEANUP(cuMemFreeHost, (ptr));
   }
 
-
-
-
-  struct host_allocation : public boost::noncopyable, public context_dependent
+#if CUDAPP_CUDA_VERSION >= 4000
+  inline void *mem_host_register(void *ptr, size_t bytes, unsigned int flags=0)
   {
-    private:
+    CUDAPP_CALL_GUARDED(cuMemHostRegister, (ptr, bytes, flags));
+    return ptr;
+  }
+
+  inline void mem_host_unregister(void *ptr)
+  {
+    CUDAPP_CALL_GUARDED_CLEANUP(cuMemHostUnregister, (ptr));
+  }
+#endif
+
+  inline void *aligned_malloc(size_t size, size_t alignment)
+  {
+    // alignment must be a power of two.
+    if ((alignment & (alignment - 1)) != 0)
+      throw pycuda::error("aligned_malloc", CUDA_ERROR_INVALID_VALUE,
+          "alignment must be a power of two");
+
+    if (alignment == 0)
+      throw pycuda::error("aligned_malloc", CUDA_ERROR_INVALID_VALUE,
+          "alignment must non-zero");
+
+    void *p = malloc(size + (alignment - 1));
+    if (!p)
+      throw pycuda::error("aligned_malloc", CUDA_ERROR_OUT_OF_MEMORY,
+          "aligned malloc failed");
+
+    return (void *)((((ptrdiff_t)(p)) + (alignment-1)) & -alignment);
+  }
+
+
+
+  struct host_pointer : public boost::noncopyable, public context_dependent
+  {
+    protected:
       bool m_valid;
       void *m_data;
 
     public:
-      host_allocation(unsigned bytesize, unsigned flags=0)
-        : m_valid(true), m_data(mem_alloc_host(bytesize, flags))
+      host_pointer()
+        : m_valid(false)
       { }
 
-      ~host_allocation()
+      host_pointer(void *ptr)
+        : m_valid(true), m_data(ptr)
+      { }
+
+      virtual ~host_pointer()
       {
         if (m_valid)
           free();
       }
 
+      virtual void free()
+      { }
+
+      void *data()
+      { return m_data; }
+
+#if CUDAPP_CUDA_VERSION >= 2020
+      CUdeviceptr get_device_pointer()
+      {
+        CUdeviceptr result;
+        CUDAPP_CALL_GUARDED(cuMemHostGetDevicePointer, (&result, m_data, 0));
+        return result;
+      }
+#endif
+
+  };
+
+  struct pagelocked_host_allocation : public host_pointer
+  {
+    public:
+      pagelocked_host_allocation(size_t bytesize, unsigned flags=0)
+        : host_pointer(mem_host_alloc(bytesize, flags))
+      { }
+
+      void free()
+      {
+        if (m_valid)
+        {
+          try
+          {
+            scoped_context_activation ca(get_context());
+            mem_host_free(m_data);
+          }
+          CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(pagelocked_host_allocation);
+
+          release_context();
+          m_valid = false;
+        }
+        else
+          throw pycuda::error("pagelocked_host_allocation::free", CUDA_ERROR_INVALID_HANDLE);
+      }
+
+#if CUDAPP_CUDA_VERSION >= 3020
+      unsigned int get_flags()
+      {
+        unsigned int flags;
+        CUDAPP_CALL_GUARDED(cuMemHostGetFlags, (&flags, m_data));
+        return flags;
+      }
+#endif
+  };
+
+  struct aligned_host_allocation : public host_pointer
+  {
+    public:
+      aligned_host_allocation(size_t size, size_t alignment)
+        : host_pointer(aligned_malloc(size, alignment))
+      { }
+
+      void free()
+      {
+        if (m_valid)
+        {
+          ::free(m_data);
+        }
+        else
+          throw pycuda::error("aligned_host_allocation::free", CUDA_ERROR_INVALID_HANDLE);
+      }
+  };
+
+#if CUDAPP_CUDA_VERSION >= 4000
+  struct registered_host_memory : public host_pointer
+  {
+    private:
+      py::object m_base;
+
+    public:
+      registered_host_memory(void *p, size_t bytes, unsigned int flags=0, 
+          py::object base=py::object())
+        : host_pointer(mem_host_register(p, bytes, flags)), m_base(base)
+      {
+      }
+
       void free()
       {
         if (m_valid)
@@ -1454,7 +1783,7 @@ namespace cuda
           try
           {
             scoped_context_activation ca(get_context());
-            mem_free_host(m_data);
+            mem_host_unregister(m_data);
           }
           CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(host_allocation);
 
@@ -1462,31 +1791,15 @@ namespace cuda
           m_valid = false;
         }
         else
-          throw cuda::error("host_allocation::free", CUDA_ERROR_INVALID_HANDLE);
+          throw pycuda::error("registered_host_memory::free", CUDA_ERROR_INVALID_HANDLE);
       }
 
-      void *data()
-      { return m_data; }
-
-#if CUDA_VERSION >= 2020
-      CUdeviceptr get_device_pointer()
+      py::object base() const
       {
-        CUdeviceptr result;
-        CUDAPP_CALL_GUARDED(cuMemHostGetDevicePointer, (&result, m_data, 0));
-        return result;
+        return m_base;
       }
-#endif
-
-#if CUDA_VERSION >= 3020
-      unsigned int get_flags()
-      {
-        unsigned int flags;
-        CUDAPP_CALL_GUARDED(cuMemHostGetFlags, (&flags, m_data));
-        return flags;
-      }
-#endif
-
   };
+#endif
 
   // }}}
 
@@ -1512,19 +1825,15 @@ namespace cuda
 
       event *record(py::object stream_py)
       {
-        CUstream s_handle;
-        if (stream_py.ptr() != Py_None)
-        {
-          const stream &s = py::extract<const stream &>(stream_py);
-          s_handle = s.handle();
-        }
-        else
-          s_handle = 0;
+        PYCUDA_PARSE_STREAM_PY;
 
         CUDAPP_CALL_GUARDED(cuEventRecord, (m_event, s_handle));
         return this;
       }
 
+      CUevent handle() const
+      { return m_event; }
+
       event *synchronize()
       {
         CUDAPP_CALL_GUARDED_THREADED(cuEventSynchronize, (m_event));
@@ -1563,6 +1872,35 @@ namespace cuda
       }
   };
 
+#if CUDAPP_CUDA_VERSION >= 3020
+  inline void stream::wait_for_event(const event &evt)
+  {
+    CUDAPP_CALL_GUARDED(cuStreamWaitEvent, (m_stream, evt.handle(), 0));
+  }
+#endif
+
+  // }}}
+
+  // {{{ profiler
+#if CUDAPP_CUDA_VERSION >= 4000
+  inline void initialize_profiler(
+      const char *config_file,
+      const char *output_file,
+      CUoutput_mode output_mode)
+  {
+    CUDAPP_CALL_GUARDED(cuProfilerInitialize, (config_file, output_file, output_mode));
+  }
+
+  inline void start_profiler()
+  {
+    CUDAPP_CALL_GUARDED(cuProfilerStart, ());
+  }
+
+  inline void stop_profiler()
+  {
+    CUDAPP_CALL_GUARDED(cuProfilerStart, ());
+  }
+#endif
   // }}}
 }
 
