@@ -19,6 +19,9 @@ static PyTypeObject PyStructType;
 /* compatibility macros */
 #if (PY_VERSION_HEX < 0x02050000)
 typedef long int Py_ssize_t;
+
+#define PyInt_FromSsize_t(x) PyInt_FromLong(x)
+#define PyInt_AsSsize_t(x) PyInt_AsLong(x)
 #endif
 
 /* If PY_STRUCT_FLOAT_COERCE is defined, the struct module will allow float
@@ -241,31 +244,7 @@ get_ulonglong(PyObject *v, unsigned PY_LONG_LONG *p)
 
 #endif
 
-/* Floating point helpers */
-
-static PyObject *
-unpack_float(const char *p,  /* start of 4-byte string */
-             int le)	     /* true for little-endian, false for big-endian */
-{
-	double x;
-
-	x = _PyFloat_Unpack4((unsigned char *)p, le);
-	if (x == -1.0 && PyErr_Occurred())
-		return NULL;
-	return PyFloat_FromDouble(x);
-}
-
-static PyObject *
-unpack_double(const char *p,  /* start of 8-byte string */
-              int le)         /* true for little-endian, false for big-endian */
-{
-	double x;
-
-	x = _PyFloat_Unpack8((unsigned char *)p, le);
-	if (x == -1.0 && PyErr_Occurred())
-		return NULL;
-	return PyFloat_FromDouble(x);
-}
+#if (SIZEOF_LONG > SIZEOF_INT)
 
 /* Helper to format the range error exceptions */
 static int
@@ -297,6 +276,7 @@ _range_error(const formatdef *f, int is_unsigned)
 	return -1;
 }
 
+#endif
 
 
 /* A large number of small routines follow, with names of the form
@@ -754,7 +734,7 @@ getentry(int c, const formatdef *f)
 
 /* Align a size according to a format code */
 
-static int
+static Py_ssize_t
 align(Py_ssize_t size, char c, const formatdef *e)
 {
 	if (e->format == c) {
@@ -1091,16 +1071,34 @@ s_pack_internal(PyStructObject *soself, PyObject *args, int offset, char* buf)
 		char *res = buf + code->offset;
 		if (e->format == 's') {
 			if (!PyString_Check(v)) {
-				PyErr_SetString(StructError,
-						"argument for 's' must "
-						"be a string");
-				return -1;
+				if (!PyObject_CheckReadBuffer(v))
+				{
+					PyErr_SetString(StructError,
+							"argument for 's' must "
+							"be a string or a buffer");
+					return -1;
+				}
+				else
+				{
+					void *buf;
+					Py_ssize_t len;
+					if (PyObject_AsReadBuffer(v, &buf, &len))
+						return -1;
+
+					if (len > code->size)
+						len = code->size;
+					if (len > 0)
+						memcpy(res, buf, len);
+				}
+			}
+			else
+			{
+				n = PyString_GET_SIZE(v);
+				if (n > code->size)
+					n = code->size;
+				if (n > 0)
+					memcpy(res, PyString_AS_STRING(v), n);
 			}
-			n = PyString_GET_SIZE(v);
-			if (n > code->size)
-				n = code->size;
-			if (n > 0)
-				memcpy(res, PyString_AS_STRING(v), n);
 		} else if (e->format == 'p') {
 			if (!PyString_Check(v)) {
 				PyErr_SetString(StructError,
